(* Ocsigen
 * http://www.ocsigen.org
 * Copyright (C) 2010 Vincent Balat
 * Copyright (C) 2011 Jérôme Vouillon, Grégoire Henry, Pierre Chambart
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, with linking exception;
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Eliom_pervasives

module JsTable = Eliommod_dom.JsTable

(* Ignore popstate before the first call to "set_content". It avoids
   loading the first page twice on a buggy Google-chrome. *)
let chrome_dummy_popstate = ref true

(* == Closure *)

let closure_table  : (poly -> Dom_html.event Js.t -> unit) JsTable.t = JsTable.create ()
let register_closure (id:int64) (f : 'a -> Dom_html.event Js.t -> unit) =
  JsTable.add closure_table ((Obj.magic id)##toString():Js.js_string Js.t)
    (Obj.magic f : poly -> Dom_html.event Js.t -> unit)
let find_closure (id:int64) : ( poly -> Dom_html.event Js.t -> unit ) Js.Optdef.t =
  JsTable.find closure_table ((Obj.magic id)##toString():Js.js_string Js.t)

(* == Process nodes (a.k.a. nodes with a unique Dom instance on each client process) *)

let (register_process_node, find_process_node) =
  let process_nodes : Dom.node Js.t JsTable.t = JsTable.create () in
  let find id =
    Js.Optdef.bind
      (JsTable.find process_nodes id)
      (fun node ->
        if Js.to_bytestring (node##nodeName##toLowerCase()) == "script" then
        (* We don't wan't to reexecute "unique" script... *)
          Js.def (Dom_html.document##createTextNode (Js.string "") :> Dom.node Js.t)
        else
          Js.def node)
  in
  let register id node = JsTable.add process_nodes id node in
  (register, find)

(* == Request nodes (a.k.a. nodes with a unique Dom instance in the current request) *)

let (register_request_node, find_request_node, reset_request_node) =
  let request_nodes : Dom.node Js.t JsTable.t ref = ref (JsTable.create ()) in
  let find id = JsTable.find !request_nodes id in
  let register id node = JsTable.add !request_nodes id node in
  let reset () = request_nodes := JsTable.create () in
  (register, find, reset)

let getElementById id =
  Js.Optdef.case (find_process_node (Js.string id))
    (fun () -> raise Not_found)
    (fun pnode -> pnode)

(* == Event *)

(* forward declaration... *)
let change_page_uri_ = ref (fun ?cookies_info href -> assert false)
let change_page_get_form_ = ref (fun ?cookies_info form href -> assert false)
let change_page_post_form_ = ref (fun ?cookies_info form href -> assert false)

let middleClick ev =
   match Dom_html.taggedEvent ev with
   | Dom_html.MouseEvent ev ->
       Dom_html.buttonPressed ev = Dom_html.Middle_button
       || Js.to_bool ev##ctrlKey
       || Js.to_bool ev##shiftKey
       || Js.to_bool ev##altKey
       || Js.to_bool ev##metaKey
   | _ -> false

let raw_a_handler node cookies_info ev =
  let href = (Js.Unsafe.coerce node : Dom_html.anchorElement Js.t)##href in
  let https = Url.get_ssl (Js.to_string href) in
  (middleClick ev)
  || (https = Some true && not Eliom_request_info.ssl_)
  || (https = Some false && Eliom_request_info.ssl_)
  || (!change_page_uri_ ?cookies_info (Js.to_string href); false)

let raw_form_handler form kind cookies_info ev =
  let action = Js.to_string form##action in
  let https = Url.get_ssl action in
  let change_page_form = match kind with
    | `Form_get -> !change_page_get_form_
    | `Form_post -> !change_page_post_form_ in
  (https = Some true && not Eliom_request_info.ssl_)
  || (https = Some false && Eliom_request_info.ssl_)
  || (change_page_form ?cookies_info form action; false)


let raw_event_handler function_id args =
  Js.Optdef.case (find_closure function_id)
    (fun () -> error "Closure not found (%Ld)" function_id)
    (fun f -> (fun ev -> try f args ev; true with False -> false))

let reify_caml_event node ce : #Dom_html.event Js.t -> bool = match ce with
  | XML.CE_call_service None -> (fun _ -> true)
  | XML.CE_call_service (Some (`A, cookies_info)) ->
      (fun ev ->
	let node = Js.Opt.get (Dom_html.CoerceTo.a node) (fun () -> error "not an anchor element") in
	raw_a_handler node cookies_info ev)
  | XML.CE_call_service (Some ((`Form_get | `Form_post) as kind, cookies_info)) ->
      (fun ev ->
	let form = Js.Opt.get (Dom_html.CoerceTo.form node) (fun () -> error "not a form element") in
	raw_form_handler form kind cookies_info ev)
  | XML.CE_client_closure f ->
      (fun ev -> try f ev; true with False -> false)
  | XML.CE_registered_closure (_, (function_id, args)) ->
      raw_event_handler function_id args

let reify_event node ev = match ev with
  | XML.Raw ev -> Js.Unsafe.variable ev
  | XML.Caml ce -> reify_caml_event node ce

let register_event_handler node acc (name, ev) =
  let f = reify_caml_event node ev in
  if name = "onload" then
    f :: acc
  else
    ( Js.Unsafe.set node (Js.bytestring name)
	(Dom_html.handler (fun ev -> Js.bool (f ev)));
      acc )

(* == Register nodes id and event in the orginal Dom. *)

let register_event_handlers acc node attribs =
  List.fold_left
    (fun acc ev -> register_event_handler (Js.Unsafe.coerce node : Dom_html.element Js.t) acc ev)
    acc
    attribs

let get_element_cookies_info elt =
  Js.Opt.to_option
    (Js.Opt.map (elt##getAttribute(Js.string Eliom_pervasives_base.RawXML.ce_call_service_attrib))
       (fun s -> of_json (Js.to_string s)))

let a_handler =
  Dom_html.full_handler
    (fun node ev ->
      let node = Js.Opt.get (Dom_html.CoerceTo.a node) (fun () -> error "not an anchor element") in
      Js.bool (raw_a_handler node (get_element_cookies_info node) ev))

let form_handler =
  Dom_html.full_handler
    (fun node ev ->
      let form = Js.Opt.get (Dom_html.CoerceTo.form node) (fun () -> error "not a form element") in
      let kind =
	if String.lowercase(Js.to_string form##_method) = "get"
	then `Form_get
	else `Form_post
      in
      Js.bool (raw_form_handler form kind (get_element_cookies_info form) ev))

let relink_process_node (node:Dom_html.element Js.t) =
  let id = Js.Opt.get
    (node##getAttribute(Js.string Eliom_pervasives_base.RawXML.node_id_attrib))
    (fun () -> error "unique node without id attribute") in
  Js.Optdef.case (find_process_node id)
    (fun () -> register_process_node id (node:>Dom.node Js.t))
    (fun pnode ->
      Js.Opt.iter (node##parentNode)
        (fun parent -> Dom.replaceChild parent pnode node))

let relink_request_node (node:Dom_html.element Js.t) =
  let id = Js.Opt.get
    (node##getAttribute(Js.string Eliom_pervasives_base.RawXML.node_id_attrib))
    (fun () -> error "unique node without id attribute") in
  Js.Optdef.case (find_request_node id)
    (fun () -> register_request_node id (node:>Dom.node Js.t))
    (fun pnode ->
      Js.Opt.iter (node##parentNode)
        (fun parent -> Dom.replaceChild parent pnode node))

let relink_request_nodes root =
  if !Eliom_config.debug_timings then
    Firebug.console##time
(Js.string "relink_request_nodes");
  Eliommod_dom.iter_nodeList
    (Eliommod_dom.select_request_nodes root)
    relink_request_node;
  if !Eliom_config.debug_timings then
    Firebug.console##timeEnd(Js.string "relink_request_nodes")

let relink_closure_node root onload table (node:Dom_html.element Js.t) =
  let aux attr =
    if attr##value##substring(0,Eliom_pervasives_base.RawXML.closure_attr_prefix_len) =
      Js.string Eliom_pervasives_base.RawXML.closure_attr_prefix
    then
      let cid = Js.to_bytestring (attr##value##substring_toEnd(
	Eliom_pervasives_base.RawXML.closure_attr_prefix_len)) in
      let (id,args) = XML.ClosureMap.find cid table in
      let closure = raw_event_handler id args in
      if attr##name = Js.string "onload" then
	(if Eliommod_dom.ancessor root node
	 (* if not inside a unique node replaced by an older one *)
	 then onload := closure :: !onload)
      else Js.Unsafe.set node (attr##name) (Dom_html.handler (fun ev -> Js.bool (closure ev)))
  in
  Eliommod_dom.iter_nodeList (node##attributes:>Dom.attr Dom.nodeList Js.t) aux

let relink_page (root:Dom_html.element Js.t) event_handlers =
  let (a_nodeList,form_nodeList,process_nodeList,closure_nodeList) =
    Eliommod_dom.select_nodes root in
  Eliommod_dom.iter_nodeList a_nodeList
    (fun node -> node##onclick <- a_handler);
  Eliommod_dom.iter_nodeList form_nodeList
    (fun node -> node##onsubmit <- form_handler);
  Eliommod_dom.iter_nodeList process_nodeList
    relink_process_node;
  let onload = ref [] in
  Eliommod_dom.iter_nodeList closure_nodeList
    (fun node -> relink_closure_node root onload event_handlers node);
  !onload

(* == Convertion from OCaml XML.elt nodes to native JavaScript Dom nodes *)

let run_load_events on_load =
  let load_evt = Eliommod_dom.createEvent (Js.string "load") in
  ignore (List.for_all (fun f -> f load_evt) on_load)

module Html5 = struct

  (* Type for partially unwrapped elt. *)
  type tmp_recontent =
    (* arguments ('econtent') are already unwrapped. *)
    | RELazy of XML.econtent Eliom_lazy.request
    | RE of XML.econtent
  type tmp_elt = {
    (* to be unwrapped *)
    tmp_elt : tmp_recontent;
    tmp_node_id : XML.node_id;
  }

  let _ =
    Eliom_unwrap.register_unwrapper
      (Eliom_unwrap.id_of_int Eliom_pervasives_base.tyxml_unwrap_id_int)
      (fun tmp_elt ->
	let elt = match tmp_elt.tmp_elt with
	  | RELazy elt -> Eliom_lazy.force elt
	  | RE elt -> elt
	in
	(* Do not rebuild dom node while unwrapping, otherwise we
	   don't have control on when "onload" event handlers are
	   triggered. *)
	match tmp_elt.tmp_node_id with
	| XML.ProcessId process_id as id ->
            Js.Optdef.case (find_process_node (Js.bytestring process_id))
              (fun () -> XML.make ~id elt)
              (fun elt -> XML.make_dom ~id elt)
	| XML.RequestId request_id as id ->
            Js.Optdef.case (find_request_node (Js.bytestring request_id))
              (fun () -> XML.make ~id elt)
              (fun elt -> XML.make_dom ~id elt)
	| XML.NoId as id -> XML.make ~id elt)

  let rebuild_attrib node name a = match a with
    | XML.AFloat f -> Js.Unsafe.set node (Js.string name) (Js.Unsafe.inject f)
    | XML.AInt i -> Js.Unsafe.set node (Js.string name) (Js.Unsafe.inject i)
    | XML.AStr s ->
      node##setAttribute(Js.string name, Js.string s)
    | XML.AStrL (XML.Space, sl) ->
      node##setAttribute(Js.string name, Js.string (String.concat " " sl))
    | XML.AStrL (XML.Comma, sl) ->
      node##setAttribute(Js.string name, Js.string (String.concat "," sl))

  let rebuild_rattrib onload_acc node ra = match XML.racontent ra with
    | XML.RA a -> rebuild_attrib node (XML.aname ra) a
    | XML.RACamlEventHandler ev ->
        onload_acc := register_event_handler node !onload_acc (XML.aname ra, ev)
    | XML.RALazyStr s ->
	node##setAttribute(Js.string (XML.aname ra), Js.string s)
    | XML.RALazyStrL (XML.Space, l) ->
	node##setAttribute(Js.string (XML.aname ra), Js.string (String.concat " " l))
    | XML.RALazyStrL (XML.Comma, l) ->
	node##setAttribute(Js.string (XML.aname ra), Js.string (String.concat "," l))

  let rec rebuild_node onload_acc elt =
    match XML.get_node elt with
    | XML.DomNode node ->
        (* assert (XML.get_node_id node <> NoId); *)
        node
    | XML.TyXMLNode raw_elt ->
        match XML.get_node_id elt with
	| XML.NoId -> raw_rebuild_node onload_acc raw_elt
	| XML.RequestId _ ->
	    (* Do not look in request_nodes hashtbl: such elements have
	       been bind while unwrapping nodes. *)
            let node = raw_rebuild_node onload_acc raw_elt in
	    XML.set_dom_node elt node;
	    node
	| XML.ProcessId id ->
	  let id = (Js.string id) in
	  Js.Optdef.case (find_process_node id)
            (fun () ->
	      let node = raw_rebuild_node onload_acc (XML.content elt) in
	      register_process_node id node;
	      node)
            (fun n -> (n:> Dom.node Js.t))


  and raw_rebuild_node onload_acc = function
    | XML.Empty
    | XML.Comment _ ->
	(* FIXME *)
	(Dom_html.document##createTextNode (Js.string "") :> Dom.node Js.t)
    | XML.EncodedPCDATA s
    | XML.PCDATA s -> (Dom_html.document##createTextNode (Js.string s) :> Dom.node Js.t)
    | XML.Entity s -> assert false (* FIXME *)
    | XML.Leaf (name,attribs) ->
      let node = Dom_html.document##createElement (Js.string name) in
      List.iter (rebuild_rattrib onload_acc node) attribs;
      (node :> Dom.node Js.t)
    | XML.Node (name,attribs,childrens) ->
      let node = Dom_html.document##createElement (Js.string name) in
      List.iter (rebuild_rattrib onload_acc node) attribs;
      List.iter (fun c -> Dom.appendChild node (rebuild_node onload_acc c)) childrens;
      (node :> Dom.node Js.t)

  let rebuild_node elt =
    let onload_acc = ref [] in
    let node = Js.Unsafe.coerce (rebuild_node onload_acc (HTML5.M.toelt elt)) in
    run_load_events !onload_acc;
    node

  let of_element = rebuild_node
  let of_heading = rebuild_node

  let of_a = rebuild_node
  let of_abbr = rebuild_node
  let of_acronym = rebuild_node
  let of_address = rebuild_node
  let of_applet = rebuild_node
  let of_area = rebuild_node
  let of_article = rebuild_node
  let of_aside = rebuild_node
  let of_audio = rebuild_node
  let of_b = rebuild_node
  let of_base = rebuild_node
  let of_basefont = rebuild_node
  let of_bdi = rebuild_node
  let of_bdo = rebuild_node
  let of_big = rebuild_node
  let of_blockquote = rebuild_node
  let of_body = rebuild_node
  let of_br = rebuild_node
  let of_button = rebuild_node
  let of_canvas = rebuild_node
  let of_caption = rebuild_node
  let of_center = rebuild_node
  let of_cite = rebuild_node
  let of_code = rebuild_node
  let of_col = rebuild_node
  let of_colgroup = rebuild_node
  let of_command = rebuild_node
  let of_datalist = rebuild_node
  let of_dd = rebuild_node
  let of_del = rebuild_node
  let of_details = rebuild_node
  let of_dfn = rebuild_node
  let of_dir = rebuild_node
  let of_div = rebuild_node
  let of_dl = rebuild_node
  let of_dt = rebuild_node
  let of_em = rebuild_node
  let of_embed = rebuild_node
  let of_fieldset = rebuild_node
  let of_figcaption = rebuild_node
  let of_figure = rebuild_node
  let of_font = rebuild_node
  let of_footer = rebuild_node
  let of_form = rebuild_node
  let of_frame = rebuild_node
  let of_frameset = rebuild_node
  let of_h1 = rebuild_node
  let of_h2 = rebuild_node
  let of_h3 = rebuild_node
  let of_h4 = rebuild_node
  let of_h5 = rebuild_node
  let of_h6 = rebuild_node
  let of_head = rebuild_node
  let of_header = rebuild_node
  let of_hgroup = rebuild_node
  let of_hr = rebuild_node
  let of_html = rebuild_node
  let of_i = rebuild_node
  let of_iframe = rebuild_node
  let of_img = rebuild_node
  let of_input = rebuild_node
  let of_ins = rebuild_node
  let of_keygen = rebuild_node
  let of_kbd = rebuild_node
  let of_label = rebuild_node
  let of_legend = rebuild_node
  let of_li = rebuild_node
  let of_link = rebuild_node
  let of_map = rebuild_node
  let of_mark = rebuild_node
  let of_menu = rebuild_node
  let of_meta = rebuild_node
  let of_meter = rebuild_node
  let of_nav = rebuild_node
  let of_noframes = rebuild_node
  let of_noscript = rebuild_node
  let of_object = rebuild_node
  let of_ol = rebuild_node
  let of_optgroup = rebuild_node
  let of_option = rebuild_node
  let of_output = rebuild_node
  let of_p = rebuild_node
  let of_param = rebuild_node
  let of_pre = rebuild_node
  let of_progress = rebuild_node
  let of_q = rebuild_node
  let of_rp = rebuild_node
  let of_rt = rebuild_node
  let of_ruby = rebuild_node
  let of_s = rebuild_node
  let of_samp = rebuild_node
  let of_script = rebuild_node
  let of_section = rebuild_node
  let of_select = rebuild_node
  let of_small = rebuild_node
  let of_source = rebuild_node
  let of_span = rebuild_node
  let of_strike = rebuild_node
  let of_strong = rebuild_node
  let of_style = rebuild_node
  let of_sub = rebuild_node
  let of_summary = rebuild_node
  let of_sup = rebuild_node
  let of_table = rebuild_node
  let of_tbody = rebuild_node
  let of_td = rebuild_node
  let of_textarea = rebuild_node
  let of_tfoot = rebuild_node
  let of_th = rebuild_node
  let of_thead = rebuild_node
  let of_time = rebuild_node
  let of_title = rebuild_node
  let of_tr = rebuild_node
  let of_track = rebuild_node
  let of_tt = rebuild_node
  let of_u = rebuild_node
  let of_ul = rebuild_node
  let of_var = rebuild_node
  let of_video = rebuild_node
  let of_wbr = rebuild_node

end

(* == XHR *)

let current_pseudo_fragment = ref ""
let url_fragment_prefix = "!"
let url_fragment_prefix_with_sharp = "#!"

let create_request_
    ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
    ?keep_nl_params ?nl_params ?keep_get_na_params
    get_params post_params =
  (* GRGR TODO: allow get_get_or_post service to return also the service
     with the correct subtype. Then do use Eliom_uri.make_string_uri
     and Eliom_uri.make_post_uri_components instead of Eliom_uri.make_string_uri_
     and Eliom_uri.make_post_uri_components__ *)
  match Eliom_services.get_get_or_post service with
    | `Get ->
        let uri =
          Eliom_uri.make_string_uri_
            ?absolute ?absolute_path ?https
            ~service
            ?hostname ?port ?fragment ?keep_nl_params ?nl_params get_params
        in
        `Get uri
    | `Post ->
        let path, get_params, fragment, post_params =
          Eliom_uri.make_post_uri_components__
            ?absolute ?absolute_path ?https
            ~service
            ?hostname ?port ?fragment ?keep_nl_params ?nl_params
            ?keep_get_na_params get_params post_params
        in
        let uri =
          Eliom_uri.make_string_uri_from_components (path, get_params, fragment)
        in
        `Post (uri, post_params)



let exit_to
    ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
    ?keep_nl_params ?nl_params ?keep_get_na_params
    get_params post_params =
  (match create_request_
     ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
     ?keep_nl_params ?nl_params ?keep_get_na_params
     get_params post_params
   with
     | `Get uri -> Eliom_request.redirect_get uri
     | `Post (uri, post_params) -> Eliom_request.redirect_post uri post_params)


(** This will change the URL, without doing a request.
    As browsers do not not allow to change the URL,
    we write the new URL in the fragment part of the URL.
    A script must do the redirection if there is something in the fragment.
    Usually this function is only for internal use.
*)

let current_uri =
  ref (fst (Url.split_fragment (Js.to_string Dom_html.window##location##href)))

let change_url_string uri =
  current_uri := fst (Url.split_fragment uri);
  if Eliom_process.history_api then begin
    Dom_html.window##history##replaceState(( Dom_html.document##body##scrollTop,
                                             Dom_html.document##body##scrollLeft),
					Js.string "" ,
					Js.null);
    Dom_html.window##history##pushState(Js.null,
					Js.string "" ,
					Js.Opt.return (Js.string uri));
  end else begin
    current_pseudo_fragment := url_fragment_prefix_with_sharp^uri;
    Eliom_request_info.set_current_path uri;
    Dom_html.window##location##hash <- Js.string (url_fragment_prefix^uri)
  end
let change_url
(*VVV is it safe to have absolute URLs? do we accept non absolute paths? *)
    ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
    ?keep_nl_params ?nl_params ?keep_get_na_params
    get_params post_params =
(*VVV only for GET services? *)
  let uri =
    (match create_request_
       ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
       ?keep_nl_params ?nl_params ?keep_get_na_params
       get_params post_params
     with
       | `Get uri -> uri
       | `Post (uri, post_params) -> uri)
  in
  change_url_string uri

let loading_phase = ref true
let load_end = Lwt_condition.create ()
let on_unload_scripts = ref []
let at_exit_scripts = ref []

(* BEGIN FORMDATA HACK: This is only needed if FormData is not available in the browser.
   When it will be commonly available, remove all sections marked by "FORMDATA HACK" !
   Notice: this hack is used to circumvent a limitation in FF4 implementation of formdata:
     if the user click on a button in a form, formdatas created in the onsubmit callback normaly contains the value of the button. ( it is the behaviour of chromium )
     in FF4, it is not the case: we must do this hack to find wich button was clicked.

   NOTICE: this may not be corrected the way we want:
     see https://bugzilla.mozilla.org/show_bug.cgi?id=647231
     html5 will explicitely specify that chromium behaviour is wrong...

   This is implemented in:
   * this file -> here and called in load_eliom_data
   * Eliom_request: in send_post_form
   * in js_of_ocaml, module Form: the code to emulate FormData *)

let onclick_on_body_handler event =
  (match Dom_html.tagged (Dom_html.eventTarget event) with
    | Dom_html.Button button ->
	(Js.Unsafe.variable "window")##eliomLastButton <- Some button;
    | Dom_html.Input input when input##_type = Js.string "submit" ->
	(Js.Unsafe.variable "window")##eliomLastButton <- Some input;
    | _ ->
	(Js.Unsafe.variable "window")##eliomLastButton <- None);
  Js._true

let add_onclick_events _ =
  ignore (Dom_html.addEventListener ( Dom_html.window##document##body )
	    Dom_html.Event.click ( Dom_html.handler onclick_on_body_handler )
	    Js._true : Dom_html.event_listener_id);
  true

(* END FORMDATA HACK *)

let broadcast_load_end _ =
  loading_phase := false;
  Lwt_condition.broadcast load_end ();
  true

let load_eliom_data js_data page =
  if !Eliom_config.debug_timings then
    Firebug.console##time(Js.string "load_eliom_data");
  loading_phase := true;
  let nodes_on_load =
    relink_page page js_data.Eliom_types.ejs_event_handler_table
  in
  Eliom_request_info.set_session_info js_data.Eliom_types.ejs_sess_info;
  let on_load =
    List.map
      (reify_caml_event Dom_html.document##documentElement)
      js_data.Eliom_types.ejs_onload
  in
  let on_unload =
    List.map
      (reify_caml_event Dom_html.document##documentElement)
      js_data.Eliom_types.ejs_onunload
  in
  let unload_evt = Eliommod_dom.createEvent (Js.string "unload") in
  on_unload_scripts :=
    [fun () -> List.for_all (fun f -> f unload_evt) on_unload];
  if !Eliom_config.debug_timings then
    Firebug.console##timeEnd(Js.string "load_eliom_data");
  add_onclick_events :: on_load @ nodes_on_load @ [broadcast_load_end]

let load_eliom_data js_data page =
  try
    load_eliom_data js_data page
  with
    | e ->
      debug_exn "load_eliom_data failed: " e;
      raise e

let in_onload () = !loading_phase

let wait_load_end () =
  if !loading_phase
  then Lwt_condition.wait load_end
  else Lwt.return ()

let on_unload f =
  on_unload_scripts :=
    (fun () -> try f(); true with False -> false) :: !on_unload_scripts

let get_data_script page =
  let head = Eliommod_dom.get_head page in
  match Dom.list_of_nodeList head##childNodes with
    | _ :: _ :: data_script :: _ ->
      let data_script = (Js.Unsafe.coerce (data_script:Dom.node Js.t):Dom.element Js.t) in
      (match Js.to_bytestring (data_script##tagName##toLowerCase ()) with
	| "script" -> (Js.Unsafe.coerce (data_script:Dom.element Js.t):Dom_html.scriptElement Js.t)
	| t ->
	  Firebug.console##error_4(Js.string "get_data_script: the node ",data_script,Js.string " is not a script, its tag is", t);
	  failwith "get_data_script")
    | _ -> error "get_data_script wrong branch"

let load_data_script data_script =
  let script = data_script##text in
  if !Eliom_config.debug_timings then
    Firebug.console##time(Js.string "load_data_script");
  ignore (Js.Unsafe.eval_string (Js.to_string script));
  if !Eliom_config.debug_timings then
    Firebug.console##timeEnd(Js.string "load_data_script");
  ( Eliom_request_info.get_request_data (),
    Eliom_request_info.get_request_cookies ())

let scroll_to_fragment ?offset fragment =
  match offset with
  | Some (scrollTop, scrollLeft) ->
    Dom_html.document##body##scrollTop <- scrollTop;
    Dom_html.document##body##scrollLeft <- scrollLeft;
  | None ->
      match fragment with
      | None | Some "" ->
          Dom_html.window##scroll(0, 0)
      | Some fragment ->
          let scroll_to_element e = e##scrollIntoView(Js._true) in
           let elem = Dom_html.document##getElementById(Js.string fragment) in
          Js.Opt.iter elem scroll_to_element

let registered_process_node id = Js.Optdef.test (find_process_node id)

let set_content ?uri ?fragment = function
  | None -> Lwt.return ()
  | Some content ->
    if !Eliom_config.debug_timings then
      Firebug.console##time(Js.string "set_content");
    try_lwt
      if !Eliom_config.debug_timings then
        Firebug.console##time(Js.string "set_content_beginning");
      chrome_dummy_popstate := false;
      ignore (List.for_all (fun f -> f ()) !on_unload_scripts);
      on_unload_scripts := [];
      (match uri, fragment with
	| Some uri, None -> change_url_string uri
	| Some uri, Some fragment ->
	  change_url_string (uri ^ "#" ^ fragment)
	| _ -> ());
      let fake_page = Eliommod_dom.html_document content registered_process_node in
      if !Eliom_config.debug_timings then
        ( Firebug.console##timeEnd(Js.string "set_content_beginning");
          Firebug.console##debug(Js.string ("loading: " ^ Js.to_string Dom_html.window##location##href)) );
      let preloaded_css = Eliommod_dom.preload_css fake_page in
      relink_request_nodes fake_page;
      let js_data, cookies = load_data_script (get_data_script fake_page) in
      Eliommod_cookies.update_cookie_table cookies;
      lwt () = preloaded_css in
      let on_load = load_eliom_data js_data fake_page in
      (* The request node table must be empty when node received
	 via call_caml_service are unwrapped. *)
      reset_request_node ();
      if !Eliom_config.debug_timings then
        Firebug.console##time(Js.string "replace_child");
      Dom.replaceChild Dom_html.document
        fake_page
	Dom_html.document##documentElement;
      if !Eliom_config.debug_timings then
        ( Firebug.console##timeEnd(Js.string "replace_child");
          Firebug.console##time(Js.string "set_content_end") );
      run_load_events on_load;
      iter_option (fun uri -> scroll_to_fragment fragment) uri;
      if !Eliom_config.debug_timings then
        ( Firebug.console##timeEnd(Js.string "set_content_end");
          Firebug.console##timeEnd(Js.string "set_content") );
      Lwt.return ()
    with
      | e ->
        debug_exn "set_content: exception raised: " e;
	if !Eliom_config.debug_timings then
          Firebug.console##timeEnd(Js.string "set_content");
        raise_lwt e

let change_page
    ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
    ?keep_nl_params ?nl_params ?keep_get_na_params
    get_params post_params =

  if not (Eliom_services.xhr_with_cookies service)
    || (https = Some true && not Eliom_request_info.ssl_)
    || (https = Some false && Eliom_request_info.ssl_)
  then
    Lwt.return
      (exit_to
         ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
	 ?keep_nl_params ?nl_params ?keep_get_na_params
         get_params post_params)
  else
    let cookies_info = Eliom_uri.make_cookies_info (https, service) in
    lwt (uri, content) = match
        create_request_
          ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
	  ?keep_nl_params ?nl_params ?keep_get_na_params
          get_params post_params
      with
	| `Get uri ->
          Eliom_request.http_get
            ~expecting_process_page:true ?cookies_info uri []
	    Eliom_request.xml_result
	| `Post (uri, p) ->
          Eliom_request.http_post
	    ~expecting_process_page:true ?cookies_info uri p
	    Eliom_request.xml_result
    in
    let uri, fragment = Url.split_fragment uri in
    set_content ~uri ?fragment content

let split_fragment uri =
  if Eliom_process.history_api then
    Url.split_fragment uri
  else
    (uri, None) (* TODO *)

let change_page_uri ?cookies_info ?(get_params = []) full_uri =
  let uri, fragment = split_fragment full_uri in
  if uri <> !current_uri || fragment = None then
    lwt (uri, content) = Eliom_request.http_get
      ~expecting_process_page:true ?cookies_info uri get_params
      Eliom_request.xml_result
    in
    set_content ~uri ?fragment content
  else
    ( change_url_string full_uri;
      scroll_to_fragment fragment;
      Lwt.return () )

let change_page_get_form ?cookies_info form full_uri =
  let form = Js.Unsafe.coerce form in
  let uri, fragment = split_fragment full_uri in
  lwt uri, content = Eliom_request.send_get_form
    ~expecting_process_page:true ?cookies_info form uri
    Eliom_request.xml_result
  in
  set_content ~uri ?fragment content

let change_page_post_form ?cookies_info form full_uri =
  let form = Js.Unsafe.coerce form in
  let uri, fragment = split_fragment full_uri in
  lwt uri, content = Eliom_request.send_post_form
      ~expecting_process_page:true ?cookies_info form uri
      Eliom_request.xml_result
  in
  set_content ~uri ?fragment content

let _ =
  change_page_uri_ :=
    (fun ?cookies_info href ->
      lwt_ignore(change_page_uri ?cookies_info href));
  change_page_get_form_ :=
    (fun ?cookies_info form href ->
      lwt_ignore(change_page_get_form ?cookies_info form href));
  change_page_post_form_ :=
    (fun ?cookies_info form href ->
      lwt_ignore(change_page_post_form ?cookies_info form href))

let call_service
    ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
    ?keep_nl_params ?nl_params ?keep_get_na_params
    get_params post_params =
  lwt _, content =
    match create_request_
      ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
      ?keep_nl_params ?nl_params ?keep_get_na_params
      get_params post_params
    with
    | `Get uri ->
	Eliom_request.http_get
          ?cookies_info:(Eliom_uri.make_cookies_info (https, service)) uri []
	  Eliom_request.string_result
    | `Post (uri, post_params) ->
	Eliom_request.http_post
          ?cookies_info:(Eliom_uri.make_cookies_info (https, service))
	  uri post_params Eliom_request.string_result in
  match content with
    | None -> raise_lwt (Eliom_request.Failed_request 204)
    | Some content -> Lwt.return content


let call_caml_service
    ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
    ?keep_nl_params ?nl_params ?keep_get_na_params
    get_params post_params =
  lwt s =
    call_service
      ?absolute ?absolute_path ?https ~service ?hostname ?port ?fragment
      ?keep_nl_params ?nl_params ?keep_get_na_params
      get_params post_params in
  let r : 'a Eliom_types.eliom_caml_service_data =
    Eliom_unwrap.unwrap (Url.decode s) 0 in
  let on_load =
    List.map
      (reify_caml_event Dom_html.document##documentElement)
      r.Eliom_types.ecs_onload in
  run_load_events on_load;
  Lwt.return r.Eliom_types.ecs_data



(*****************************************************************************)
(* Make the back button work when only the fragment has changed ... *)
(*VVV We check the fragment every t second ... :-( *)

let write_fragment s = Dom_html.window##location##hash <- Js.string s

let read_fragment () = Js.to_string Dom_html.window##location##hash

let auto_change_page fragment =
  lwt_ignore
    (let l = String.length fragment in
     if (l = 0) || ((l > 1) && (fragment.[1] = '!'))
     then
       if fragment <> !current_pseudo_fragment
       then
         (
           current_pseudo_fragment := fragment;
           let uri =
	     match l with
	       | 2 -> "./" (* fix for firefox *)
	       | 0 | 1 -> Url.Current.as_string
	       | _ -> String.sub fragment 2 ((String.length fragment) - 2)
           in
           lwt (_, content) =
	     Eliom_request.http_get ~expecting_process_page:true uri []
	       Eliom_request.xml_result in
	   set_content content
	 )
       else Lwt.return ()
     else Lwt.return ())


let _ =

  if Eliom_process.history_api then

    Dom_html.window##onpopstate <-
      Dom_html.handler
      (fun event ->
	let full_uri = Js.to_string Dom_html.window##location##href in
        Dom_html.window##history##replaceState(( Dom_html.document##body##scrollTop,
                                                 Dom_html.document##body##scrollLeft),
					       Js.string "" ,
					       Js.null);
        let offset = Js.Opt.to_option (Obj.magic event##state : (int * int) Js.opt) in
	lwt_ignore
	  (let uri, fragment = split_fragment full_uri in
	   if uri <> !current_uri then
	     lwt uri, content =
	       Eliom_request.http_get ~expecting_process_page:true uri []
		 Eliom_request.xml_result in
	     current_uri := uri;
	     set_content content >>
	       ( scroll_to_fragment ?offset fragment;
		 Lwt.return ())
	   else
	     ( scroll_to_fragment ?offset fragment;
	       Lwt.return () ));
	Js._false)

  else

    let (fragment, set_fragment_signal) = React.S.create (read_fragment ()) in

    let rec fragment_polling () =
      lwt () = Lwt_js.sleep 0.2 in
      let new_fragment = read_fragment () in
      set_fragment_signal new_fragment;
      fragment_polling () in

    ignore(fragment_polling ());
    ignore(React.E.map auto_change_page (React.S.changes fragment))
